# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from dataclasses import dataclass, field
import json
from pathlib import Path
import re
from typing import AsyncIterator

from swh.fuse.fs.artifact import OBJTYPE_GETTERS, SWHID_REGEXP, Origin
from swh.fuse.fs.entry import (
    EntryMode,
    FuseDirEntry,
    FuseEntry,
    FuseFileEntry,
    FuseSymlinkEntry,
)
from swh.model.exceptions import ValidationError
from swh.model.identifiers import CONTENT, SWHID, parse_swhid

JSON_SUFFIX = ".json"


@dataclass
class Root(FuseDirEntry):
    """ The FUSE mountpoint, consisting of the archive/ and origin/ directories """

    name: str = field(init=False, default=None)
    mode: int = field(init=False, default=int(EntryMode.RDONLY_DIR))
    depth: int = field(init=False, default=1)

    async def compute_entries(self) -> AsyncIterator[FuseEntry]:
        yield self.create_child(ArchiveDir)
        yield self.create_child(OriginDir)
        yield self.create_child(CacheDir)


@dataclass
class ArchiveDir(FuseDirEntry):
    """ The `archive/` virtual directory allows to mount any artifact on the fly
    using its SWHID as name. The associated metadata of the artifact from the
    Software Heritage Web API can also be accessed through the `SWHID.json` file
    (in case of pagination, the JSON file will contain a complete version with
    all pages merged together). Note: the archive directory cannot be listed
    with ls, but entries in it can be accessed (e.g., using cat or cd). """

    name: str = field(init=False, default="archive")
    mode: int = field(init=False, default=int(EntryMode.RDONLY_DIR))

    ENTRIES_REGEXP = re.compile(r"^(" + SWHID_REGEXP + ")(.json)?$")

    async def compute_entries(self) -> AsyncIterator[FuseEntry]:
        return
        yield

    async def lookup(self, name: str) -> FuseEntry:
        # On the fly mounting of a new artifact
        try:
            if name.endswith(JSON_SUFFIX):
                swhid = parse_swhid(name[: -len(JSON_SUFFIX)])
                return self.create_child(
                    MetaEntry,
                    name=f"{swhid}{JSON_SUFFIX}",
                    mode=int(EntryMode.RDONLY_FILE),
                    swhid=swhid,
                )
            else:
                swhid = parse_swhid(name)
                await self.fuse.get_metadata(swhid)
                return self.create_child(
                    OBJTYPE_GETTERS[swhid.object_type],
                    name=str(swhid),
                    mode=int(
                        EntryMode.RDONLY_FILE
                        if swhid.object_type == CONTENT
                        else EntryMode.RDONLY_DIR
                    ),
                    swhid=swhid,
                )
        except ValidationError:
            return None


@dataclass
class MetaEntry(FuseFileEntry):
    """ An entry for a `archive/<SWHID>.json` file, containing all the SWHID's
    metadata from the Software Heritage archive. """

    swhid: SWHID

    async def get_content(self) -> bytes:
        # Make sure the metadata is in cache
        await self.fuse.get_metadata(self.swhid)
        # Retrieve raw JSON metadata from cache (un-typified)
        metadata = await self.fuse.cache.metadata.get(self.swhid, typify=False)
        json_str = json.dumps(metadata, indent=self.fuse.conf["json-indent"])
        return (json_str + "\n").encode()

    async def size(self) -> int:
        return len(await self.get_content())


@dataclass
class OriginDir(FuseDirEntry):
    """ The origin/ directory is lazily populated with one entry per accessed
    origin URL (mangled to create a valid UNIX filename). The URL encoding is
    done using the percent-encoding mechanism described in RFC 3986. """

    name: str = field(init=False, default="origin")
    mode: int = field(init=False, default=int(EntryMode.RDONLY_DIR))

    ENTRIES_REGEXP = re.compile(r"^.*%3A.*$")  # %3A is the encoded version of ':'

    def create_child(self, url_encoded: str) -> FuseEntry:
        return super().create_child(
            Origin, name=url_encoded, mode=int(EntryMode.RDONLY_DIR),
        )

    async def compute_entries(self) -> AsyncIterator[FuseEntry]:
        async for url in self.fuse.cache.get_cached_visits():
            yield self.create_child(url)

    async def lookup(self, name: str) -> FuseEntry:
        entry = await super().lookup(name)
        if entry:
            return entry

        # On the fly mounting of new origin url
        try:
            url_encoded = name
            await self.fuse.get_visits(url_encoded)
            return self.create_child(url_encoded)
        except ValueError:
            return None


@dataclass
class CacheDir(FuseDirEntry):
    """ The cache/ directory is an on-disk representation of locally cached
    objects and metadata. Via this directory you can browse cached data and
    selectively remove them from the cache, freeing disk space. (See `swh fs
    clean` in the {ref}`CLI <swh-fuse-cli>` to completely empty the cache). The
    directory is populated with symlinks to: all artifacts, identified by their
    SWHIDs and sharded by the first two character of their object id, the
    metadata identified by a `SWHID.json` entry, and the `origin/` directory.
    """

    name: str = field(init=False, default="cache")
    mode: int = field(init=False, default=int(EntryMode.RDONLY_DIR))

    ENTRIES_REGEXP = re.compile(r"^([a-f0-9]{2})|(" + OriginDir.name + ")$")

    @dataclass
    class ArtifactShardBySwhid(FuseDirEntry):
        ENTRIES_REGEXP = re.compile(r"^(" + SWHID_REGEXP + ")$")

        prefix: str = field(default="")

        async def compute_entries(self) -> AsyncIterator[FuseEntry]:
            root_path = self.get_relative_root_path()
            async for swhid in self.fuse.cache.get_cached_swhids():
                if not swhid.object_id.startswith(self.prefix):
                    continue

                yield self.create_child(
                    FuseSymlinkEntry,
                    name=str(swhid),
                    target=Path(root_path, f"archive/{swhid}"),
                )
                yield self.create_child(
                    FuseSymlinkEntry,
                    name=f"{swhid}{JSON_SUFFIX}",
                    target=Path(root_path, f"archive/{swhid}{JSON_SUFFIX}"),
                )

    async def compute_entries(self) -> AsyncIterator[FuseEntry]:
        prefixes = set()
        async for swhid in self.fuse.cache.get_cached_swhids():
            prefixes.add(swhid.object_id[:2])

        for prefix in prefixes:
            yield self.create_child(
                CacheDir.ArtifactShardBySwhid,
                name=prefix,
                mode=int(EntryMode.RDONLY_DIR),
                prefix=prefix,
            )

        yield self.create_child(
            FuseSymlinkEntry,
            name=OriginDir.name,
            target=Path(self.get_relative_root_path(), OriginDir.name),
        )



from swh.fuse.tests.common import get_data_from_web_archive
from swh.fuse.tests.data.config import REGULAR_FILE


def test_access_file(fuse_mntdir):
    file_path = fuse_mntdir / "archive" / REGULAR_FILE
    assert file_path.is_file()


def test_cat_file(fuse_mntdir):
    file_path = fuse_mntdir / "archive" / REGULAR_FILE
    expected = get_data_from_web_archive(REGULAR_FILE, raw=True)
    assert file_path.read_text() == expected



import json
import os
import time

import dateutil.parser

from swh.fuse.fs.artifact import RevisionHistoryShardByDate, RevisionHistoryShardByPage
from swh.fuse.tests.api_url import GRAPH_API_REQUEST
from swh.fuse.tests.common import (
    check_dir_name_entries,
    get_data_from_graph_archive,
    get_data_from_web_archive,
)
from swh.fuse.tests.data.config import REV_SMALL_HISTORY, ROOT_DIR, ROOT_REV
from swh.model.identifiers import parse_swhid


def test_access_meta(fuse_mntdir):
    file_path = fuse_mntdir / "archive" / ROOT_REV / "meta.json"
    expected = json.dumps(get_data_from_web_archive(ROOT_REV))
    assert file_path.read_text().strip() == expected.strip()


def test_list_root(fuse_mntdir):
    dir_path = fuse_mntdir / "archive" / ROOT_REV / "root"
    check_dir_name_entries(dir_path, ROOT_DIR)


def test_list_parents(fuse_mntdir):
    rev_meta = get_data_from_web_archive(ROOT_REV)
    dir_path = fuse_mntdir / "archive" / ROOT_REV / "parents"
    for i, parent in enumerate(rev_meta["parents"]):
        parent_path = dir_path / str(i + 1)
        parent_swhid = f"swh:1:rev:{parent['id']}"
        assert parent_path.is_symlink()
        assert os.readlink(parent_path) == f"../../../archive/{parent_swhid}"


def test_list_parent(fuse_mntdir):
    file_path = fuse_mntdir / "archive" / ROOT_REV / "parent"
    assert file_path.is_symlink()
    assert os.readlink(file_path) == "parents/1/"


def test_list_history(fuse_mntdir):
    dir_path = fuse_mntdir / "archive" / REV_SMALL_HISTORY / "history"
    assert os.listdir(dir_path) == ["by-date", "by-hash", "by-page"]

    history_meta = get_data_from_graph_archive(
        REV_SMALL_HISTORY, GRAPH_API_REQUEST.HISTORY
    )
    history = history_meta.strip()
    # Only keep second node in the edge because first node is redundant
    # information or the root node (hence not an ancestor)
    expected = set(
        map(parse_swhid, [edge.split(" ")[1] for edge in history.split("\n")])
    )

    dir_by_hash = dir_path / "by-hash"
    for swhid in expected:
        depth1 = swhid.object_id[:2]
        depth2 = str(swhid)
        assert (dir_by_hash / depth1).exists()
        assert depth2 in (os.listdir(dir_by_hash / depth1))

    dir_by_page = dir_path / "by-page"
    for idx, swhid in enumerate(expected):
        page_number = idx // RevisionHistoryShardByPage.PAGE_SIZE
        depth1 = RevisionHistoryShardByPage.PAGE_FMT.format(page_number=page_number)
        depth2 = str(swhid)
        assert (dir_by_page / depth1).exists()
        assert depth2 in (os.listdir(dir_by_page / depth1))

    dir_by_date = dir_path / "by-date"
    # TODO: rely on .status file instead to wait
    # Wait 2 seconds to populate by-date/ dir
    time.sleep(2)
    for swhid in expected:
        meta = get_data_from_web_archive(str(swhid))
        date = dateutil.parser.parse(meta["date"])
        depth1 = RevisionHistoryShardByDate.DATE_FMT.format(
            year=date.year, month=date.month, day=date.day
        )
        depth2 = str(swhid)
        assert (dir_by_date / depth1).exists()
        assert depth2 in (os.listdir(dir_by_date / depth1))
    assert not (dir_by_date / ".status").exists()



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from click.testing import CliRunner
import yaml

import swh.fuse.cli as cli

# mount/umount commands are already tested when setting up the fuse_mntdir fixture


def test_clean_command(tmp_path):
    fake_metadata_db = tmp_path / "metadata.sqlite"
    fake_blob_db = tmp_path / "blob.sqlite"
    config_path = tmp_path / "config.yml"
    config = {
        "cache": {
            "metadata": {"path": str(fake_metadata_db)},
            "blob": {"path": str(fake_blob_db)},
        },
    }

    fake_metadata_db.touch()
    fake_blob_db.touch()
    config_path.write_text(yaml.dump(config))

    assert fake_metadata_db.exists()
    assert fake_blob_db.exists()

    CliRunner().invoke(
        cli.fuse, args=["--config-file", str(config_path), "clean",],
    )

    assert not fake_metadata_db.exists()
    assert not fake_blob_db.exists()



import json

from swh.fuse.tests.common import get_data_from_web_archive
from swh.fuse.tests.data.config import ALL_ENTRIES


def test_access_meta_file(fuse_mntdir):
    for swhid in ALL_ENTRIES:
        # On the fly mounting
        file_path_meta = fuse_mntdir / f"archive/{swhid}.json"
        expected = json.dumps(get_data_from_web_archive(swhid))
        assert file_path_meta.read_text().strip() == expected.strip()



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from swh.fuse.tests.data.config import ORIGIN_URL_ENCODED, REGULAR_FILE
from swh.model.identifiers import parse_swhid


def test_mountpoint(fuse_mntdir):
    assert {"archive", "cache", "origin"} <= set(os.listdir(fuse_mntdir))


def test_on_the_fly_mounting(fuse_mntdir):
    assert os.listdir(fuse_mntdir / "archive") == []
    assert (fuse_mntdir / "archive" / REGULAR_FILE).is_file()
    assert (fuse_mntdir / "archive" / (REGULAR_FILE + ".json")).is_file()

    assert os.listdir(fuse_mntdir / "origin") == []
    assert (fuse_mntdir / "origin" / ORIGIN_URL_ENCODED).is_dir()

    sharded_dir = parse_swhid(REGULAR_FILE).object_id[:2]
    assert os.listdir(fuse_mntdir / "cache") == [sharded_dir, "origin"]
    assert os.listdir(fuse_mntdir / "cache" / sharded_dir) == [
        REGULAR_FILE,
        REGULAR_FILE + ".json",
    ]
    assert os.listdir(fuse_mntdir / "cache/origin") == [ORIGIN_URL_ENCODED]



import json
import os

from swh.fuse.tests.common import check_dir_name_entries, get_data_from_web_archive
from swh.fuse.tests.data.config import (
    REL_TARGET_CNT,
    REL_TARGET_DIR,
    ROOT_DIR,
    ROOT_REL,
    TARGET_CNT,
    TARGET_DIR,
)


def test_access_meta(fuse_mntdir):
    file_path = fuse_mntdir / "archive" / ROOT_REL / "meta.json"
    expected = json.dumps(get_data_from_web_archive(ROOT_REL))
    assert file_path.read_text().strip() == expected.strip()


def test_access_rev_target(fuse_mntdir):
    target_path = fuse_mntdir / "archive" / ROOT_REL / "target"
    expected = set(["meta.json", "root", "parent", "parents", "history"])
    actual = set(os.listdir(target_path))
    assert expected.issubset(actual)


def test_access_dir_target(fuse_mntdir):
    target_path = fuse_mntdir / "archive" / REL_TARGET_DIR / "target"
    check_dir_name_entries(target_path, TARGET_DIR)


def test_access_cnt_target(fuse_mntdir):
    target_path = fuse_mntdir / "archive" / REL_TARGET_CNT / "target"
    expected = get_data_from_web_archive(TARGET_CNT, raw=True)
    assert target_path.read_text() == expected


def test_target_type(fuse_mntdir):
    file_path = fuse_mntdir / "archive" / ROOT_REL / "target_type"
    assert file_path.read_text() == "revision\n"


def test_access_root(fuse_mntdir):
    dir_path = fuse_mntdir / "archive" / ROOT_REL / "root"
    check_dir_name_entries(dir_path, ROOT_DIR)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import errno
import functools
import logging
import os
from pathlib import Path
import time
from typing import Any, Dict, List
import urllib.parse

import pyfuse3
import pyfuse3_asyncio
import requests

from swh.fuse import LOGGER_NAME
from swh.fuse.cache import FuseCache
from swh.fuse.fs.entry import FuseDirEntry, FuseEntry, FuseFileEntry, FuseSymlinkEntry
from swh.fuse.fs.mountpoint import Root
from swh.model.identifiers import CONTENT, REVISION, SWHID
from swh.web.client.client import WebAPIClient


class Fuse(pyfuse3.Operations):
    """ Software Heritage Filesystem in Userspace (FUSE). Locally mount parts of
    the archive and navigate it as a virtual file system. """

    def __init__(
        self, root_path: Path, cache: FuseCache, conf: Dict[str, Any],
    ):
        super(Fuse, self).__init__()

        self._next_inode: int = pyfuse3.ROOT_INODE
        self._inode2entry: Dict[int, FuseEntry] = {}

        self.root = Root(fuse=self)
        self.conf = conf
        self.logger = logging.getLogger(LOGGER_NAME)

        self.time_ns: int = time.time_ns()  # start time, used as timestamp
        self.gid = os.getgid()
        self.uid = os.getuid()

        self.web_api = WebAPIClient(
            conf["web-api"]["url"], conf["web-api"]["auth-token"]
        )
        self.cache = cache

    def shutdown(self) -> None:
        pass

    def _alloc_inode(self, entry: FuseEntry) -> int:
        """ Return a unique inode integer for a given entry """

        inode = self._next_inode
        self._next_inode += 1
        self._inode2entry[inode] = entry

        # TODO add inode recycling with invocation to invalidate_inode when
        # the dicts get too big

        return inode

    def _remove_inode(self, inode: int) -> None:
        try:
            del self._inode2entry[inode]
        except KeyError:
            pass

    def inode2entry(self, inode: int) -> FuseEntry:
        """ Return the entry matching a given inode """

        try:
            return self._inode2entry[inode]
        except KeyError:
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def get_metadata(self, swhid: SWHID) -> Any:
        """ Retrieve metadata for a given SWHID using Software Heritage API """

        cache = await self.cache.metadata.get(swhid)
        if cache:
            return cache

        try:
            typify = False  # Get the raw JSON from the API
            # TODO: async web API
            loop = asyncio.get_event_loop()
            metadata = await loop.run_in_executor(None, self.web_api.get, swhid, typify)
            await self.cache.metadata.set(swhid, metadata)
            # Retrieve it from cache so it is correctly typed
            return await self.cache.metadata.get(swhid)
        except requests.HTTPError as err:
            self.logger.error("Cannot fetch metadata for object %s: %s", swhid, err)
            raise

    async def get_blob(self, swhid: SWHID) -> bytes:
        """ Retrieve the blob bytes for a given content SWHID using Software
        Heritage API """

        if swhid.object_type != CONTENT:
            raise pyfuse3.FUSEError(errno.EINVAL)

        # Make sure the metadata cache is also populated with the given SWHID
        await self.get_metadata(swhid)

        cache = await self.cache.blob.get(swhid)
        if cache:
            self.logger.debug("Found blob %s in cache", swhid)
            return cache

        try:
            self.logger.debug("Retrieving blob %s via web API...", swhid)
            loop = asyncio.get_event_loop()
            resp = await loop.run_in_executor(None, self.web_api.content_raw, swhid)
            blob = b"".join(list(resp))
            await self.cache.blob.set(swhid, blob)
            return blob
        except requests.HTTPError as err:
            self.logger.error("Cannot fetch blob for object %s: %s", swhid, err)
            raise

    async def get_history(self, swhid: SWHID) -> List[SWHID]:
        """ Retrieve a revision's history using Software Heritage Graph API """

        if swhid.object_type != REVISION:
            raise pyfuse3.FUSEError(errno.EINVAL)

        cache = await self.cache.history.get(swhid)
        if cache:
            self.logger.debug(
                "Found history of %s in cache (%d ancestors)", swhid, len(cache)
            )
            return cache

        try:
            # Use the swh-graph API to retrieve the full history very fast
            self.logger.debug("Retrieving history of %s via graph API...", swhid)
            call = f"graph/visit/edges/{swhid}?edges=rev:rev"
            loop = asyncio.get_event_loop()
            history = await loop.run_in_executor(None, self.web_api._call, call)
            await self.cache.history.set(history.text)
            # Retrieve it from cache so it is correctly typed
            res = await self.cache.history.get(swhid)
            return res
        except requests.HTTPError as err:
            self.logger.error("Cannot fetch history for object %s: %s", swhid, err)
            # Ignore exception since swh-graph does not necessarily contain the
            # most recent artifacts from the archive. Computing the full history
            # from the Web API is too computationally intensive so simply return
            # an empty list.
            return []

    async def get_visits(self, url_encoded: str) -> List[Dict[str, Any]]:
        """ Retrieve origin visits given an encoded-URL using Software Heritage API """

        cache = await self.cache.metadata.get_visits(url_encoded)
        if cache:
            self.logger.debug(
                "Found %d visits for origin '%s' in cache", len(cache), url_encoded,
            )
            return cache

        try:
            self.logger.debug(
                "Retrieving visits for origin '%s' via web API...", url_encoded
            )
            typify = False  # Get the raw JSON from the API
            loop = asyncio.get_event_loop()
            # Web API only takes non-encoded URL
            url = urllib.parse.unquote_plus(url_encoded)

            origin_exists = await loop.run_in_executor(
                None, self.web_api.origin_exists, url
            )
            if not origin_exists:
                raise ValueError("origin does not exist")

            visits_it = await loop.run_in_executor(
                None, functools.partial(self.web_api.visits, url, typify=typify)
            )
            visits = list(visits_it)
            await self.cache.metadata.set_visits(url_encoded, visits)
            # Retrieve it from cache so it is correctly typed
            res = await self.cache.metadata.get_visits(url_encoded)
            return res
        except (ValueError, requests.HTTPError) as err:
            self.logger.error(
                "Cannot fetch visits for origin '%s': %s", url_encoded, err
            )
            raise

    async def get_attrs(self, entry: FuseEntry) -> pyfuse3.EntryAttributes:
        """ Return entry attributes """

        attrs = pyfuse3.EntryAttributes()
        attrs.st_size = 0
        attrs.st_atime_ns = self.time_ns
        attrs.st_ctime_ns = self.time_ns
        attrs.st_mtime_ns = self.time_ns
        attrs.st_gid = self.gid
        attrs.st_uid = self.uid
        attrs.st_ino = entry.inode
        attrs.st_mode = entry.mode
        attrs.st_size = await entry.size()
        return attrs

    async def getattr(
        self, inode: int, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.EntryAttributes:
        """ Get attributes for a given inode """

        entry = self.inode2entry(inode)
        return await self.get_attrs(entry)

    async def opendir(self, inode: int, _ctx: pyfuse3.RequestContext) -> int:
        """ Open a directory referred by a given inode """

        # Re-use inode as directory handle
        self.logger.debug("opendir(inode=%d)", inode)
        return inode

    async def readdir(self, fh: int, offset: int, token: pyfuse3.ReaddirToken) -> None:
        """ Read entries in an open directory """

        # opendir() uses inode as directory handle
        inode = fh
        direntry = self.inode2entry(inode)
        self.logger.debug(
            "readdir(dirname=%s, fh=%d, offset=%d)", direntry.name, fh, offset
        )
        assert isinstance(direntry, FuseDirEntry)

        next_id = offset + 1
        try:
            async for entry in direntry.get_entries(offset):
                name = os.fsencode(entry.name)
                attrs = await self.get_attrs(entry)
                if not pyfuse3.readdir_reply(token, name, attrs, next_id):
                    break

                next_id += 1
                self._inode2entry[attrs.st_ino] = entry
        except Exception as err:
            self.logger.exception("Cannot readdir: %s", err)
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def open(
        self, inode: int, _flags: int, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.FileInfo:
        """ Open an inode and return a unique file handle """

        # Re-use inode as file handle
        self.logger.debug("open(inode=%d)", inode)
        entry = self.inode2entry(inode)
        return pyfuse3.FileInfo(fh=inode, **entry.file_info_attrs)

    async def read(self, fh: int, offset: int, length: int) -> bytes:
        """ Read `length` bytes from file handle `fh` at position `offset` """

        # open() uses inode as file handle
        inode = fh

        entry = self.inode2entry(inode)
        self.logger.debug(
            "read(name=%s, fh=%d, offset=%d, length=%d)", entry.name, fh, offset, length
        )
        assert isinstance(entry, FuseFileEntry)

        try:
            data = await entry.get_content()
            return data[offset : offset + length]
        except Exception as err:
            self.logger.exception("Cannot read: %s", err)
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def lookup(
        self, parent_inode: int, name: str, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.EntryAttributes:
        """ Look up a directory entry by name and get its attributes """

        name = os.fsdecode(name)
        parent_entry = self.inode2entry(parent_inode)
        self.logger.debug(
            "lookup(parent_name=%s, parent_inode=%d, name=%s)",
            parent_entry.name,
            parent_inode,
            name,
        )
        assert isinstance(parent_entry, FuseDirEntry)

        try:
            if parent_entry.validate_entry(name):
                lookup_entry = await parent_entry.lookup(name)
                if lookup_entry:
                    return await self.get_attrs(lookup_entry)
        except Exception as err:
            self.logger.exception("Cannot lookup: %s", err)

        raise pyfuse3.FUSEError(errno.ENOENT)

    async def readlink(self, inode: int, _ctx: pyfuse3.RequestContext) -> bytes:
        entry = self.inode2entry(inode)
        self.logger.debug("readlink(name=%s, inode=%d)", entry.name, inode)
        assert isinstance(entry, FuseSymlinkEntry)
        return os.fsencode(entry.get_target())


async def main(swhids: List[SWHID], root_path: Path, conf: Dict[str, Any]) -> None:
    """ swh-fuse CLI entry-point """

    # Use pyfuse3 asyncio layer to match the rest of Software Heritage codebase
    pyfuse3_asyncio.enable()

    async with FuseCache(conf["cache"]) as cache:
        fs = Fuse(root_path, cache, conf)

        # Initially populate the cache
        for swhid in swhids:
            try:
                await fs.get_metadata(swhid)
            except Exception as err:
                fs.logger.exception("Cannot prefetch object %s: %s", swhid, err)

        fuse_options = set(pyfuse3.default_options)
        fuse_options.add("fsname=swhfs")

        try:
            pyfuse3.init(fs, root_path, fuse_options)
            await pyfuse3.main()
        except Exception as err:
            fs.logger.error("Error running FUSE: %s", err)
        finally:
            fs.shutdown()
            pyfuse3.close(unmount=True)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from multiprocessing import Process
import os
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
import time

from click.testing import CliRunner
import pytest
import yaml

import swh.fuse.cli as cli
from swh.fuse.tests.data.api_data import API_URL, MOCK_ARCHIVE


@pytest.fixture
def web_api_mock(requests_mock):
    for api_call, data in MOCK_ARCHIVE.items():
        # Convert Python dict JSON into a string (only for non-raw API call)
        if not api_call.endswith("raw/") and not api_call.startswith("graph/"):
            data = json.dumps(data)

        http_method = requests_mock.get
        if api_call.startswith("origin/") and api_call.endswith("get/"):
            http_method = requests_mock.head

        http_method(f"{API_URL}/{api_call}", text=data)

    return requests_mock


@pytest.fixture
def fuse_mntdir(web_api_mock):
    tmpdir = TemporaryDirectory(suffix=".swh-fuse-test")

    config = {
        "cache": {"metadata": {"in-memory": True}, "blob": {"in-memory": True},},
        "web-api": {"url": API_URL, "auth-token": None},
        "json-indent": None,
    }

    # Run FUSE in foreground mode but in a separate process, so it does not
    # block execution and remains easy to kill during teardown
    def fuse_process(mntdir: Path):
        with NamedTemporaryFile(suffix=".swh-fuse-test.yml") as tmpfile:
            config_path = Path(tmpfile.name)
            config_path.write_text(yaml.dump(config))
            CliRunner().invoke(
                cli.fuse,
                args=[
                    "--config-file",
                    str(config_path),
                    "mount",
                    str(mntdir),
                    "--foreground",
                ],
            )

    fuse = Process(target=fuse_process, args=[Path(tmpdir.name)])
    fuse.start()
    # Wait max 3 seconds for the FUSE to correctly mount
    for i in range(30):
        try:
            root = os.listdir(tmpdir.name)
            if root:
                break
        except FileNotFoundError:
            pass
        time.sleep(0.1)
    else:
        raise FileNotFoundError(f"Could not mount FUSE in {tmpdir.name}")

    yield Path(tmpdir.name)

    CliRunner().invoke(cli.umount, [tmpdir.name])
    fuse.join()



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from enum import Enum
from typing import Union

from swh.model.identifiers import (
    CONTENT,
    DIRECTORY,
    RELEASE,
    REVISION,
    SNAPSHOT,
    SWHID,
    parse_swhid,
)

GRAPH_API_REQUEST = Enum("GRAPH_API_REQUEST", "HISTORY")


def swhid_to_web_url(swhid: Union[SWHID, str], raw: bool = False) -> str:
    if isinstance(swhid, str):
        swhid = parse_swhid(swhid)

    prefix = {
        CONTENT: "content/sha1_git:",
        DIRECTORY: "directory/",
        REVISION: "revision/",
        RELEASE: "release/",
        SNAPSHOT: "snapshot/",
    }

    url = f"{prefix[swhid.object_type]}{swhid.object_id}/"
    if raw:
        url += "raw/"
    return url


def swhid_to_graph_url(
    swhid: Union[SWHID, str], request_type: GRAPH_API_REQUEST
) -> str:
    if isinstance(swhid, str):
        swhid = parse_swhid(swhid)

    prefix = {
        GRAPH_API_REQUEST.HISTORY: "graph/visit/edges/",
    }

    return f"{prefix[request_type]}{swhid}"



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

""" Use the Rust compiler (v1.42.0) as a testing repository """


def remove_swhid_prefix(swhid_str: str) -> str:
    prefix = "swh:1:XXX:"
    return swhid_str[len(prefix) :]


# Content
REGULAR_FILE = "swh:1:cnt:61d3c9e1157203f0c4ed5165608d92294eaca808"
# Directory
ROOT_DIR = "swh:1:dir:c6dcbe9711ea6d5a31429a833a3d0c59cbbb2578"
# Directory (symlinks entries)
CNT_SYMLINK = "swh:1:cnt:76219eb72e8524f15c21ec93b9b2592da49b5460"  # from rust-clippy
DIR_WITH_CNT_SYMLINK = "swh:1:dir:8f04a97403e13995c8790aef9158c8981b026223"
DIR_SYMLINK = "swh:1:cnt:e8310385c56dc4bbe379f43400f3181f6a59f260"  # from Limnoria
DIR_WITH_DIR_SYMLINK = "swh:1:dir:0653a2af52ebb771ef8ce0388c942d77f9d7aee8"
REV_SYMLINKS = [  # from rust compiler
    "swh:1:rev:87dd6843678575f8dda962f239d14ef4be14b352",
    "swh:1:rev:1a2390247ad6d08160e0dd74f40a01a9578659c2",
    "swh:1:rev:4d78994915af1bde9a95c04a8c27d8dca066232a",
    "swh:1:rev:3e6e1001dc6e095dbd5c88005e80969f60e384e1",
    "swh:1:rev:11e893fc1357bc688418ddf1087c2b7aa25d154d",
    "swh:1:rev:1c2bd024d13f8011307e13386cf1fea2180352b5",
    "swh:1:rev:92baf7293dd2d418d2ac4b141b0faa822075d9f7",
]
DIR_WITH_REV_SYMLINK = "swh:1:dir:80ae84abc6122c47aae597fde99645f8663d1aba"
# Revision
ROOT_REV = "swh:1:rev:b8cedc00407a4c56a3bda1ed605c6fc166655447"
REV_SMALL_HISTORY = "swh:1:rev:37426e42cf78a43779312d780eecb21a64006d99"
# Release
ROOT_REL = "swh:1:rel:874f7cbe352033cac5a8bc889847da2fe1d13e9f"
# Snapshot
# WARNING: Do not use a snapshot artifact which is paginated because it will be
# a pain to synchronize formats properly between the mock Web API/Web API Client
# (since they differ slightly for snapshots). The Web API Client is already
# responsible for merging everything together so the real API will have no
# problem, only the mock offline one.
ROOT_SNP = "swh:1:snp:02db117fef22434f1658b833a756775ca6effed0"
ROOT_SNP_MASTER_BRANCH = "swh:1:rev:430a9fd4c797c50cea26157141b2408073b2ed91"
FAKE_SNP_SPECIAL_CASES_SWHID = "swh:1:snp:0000000000000000000000000000000000000000"
FAKE_SNP_SPECIAL_CASES = {
    # All possible target types
    "mycnt": {"target_type": "content", "target": remove_swhid_prefix(REGULAR_FILE),},
    "mydir": {"target_type": "directory", "target": remove_swhid_prefix(ROOT_DIR),},
    "myrev": {"target_type": "revision", "target": remove_swhid_prefix(ROOT_REV),},
    "myrel": {"target_type": "release", "target": remove_swhid_prefix(ROOT_REL),},
    "refs/heads/master": {
        "target_type": "revision",
        "target": remove_swhid_prefix(ROOT_SNP_MASTER_BRANCH),
    },
    # Alias with different target paths
    "alias-rootdir": {
        "target_type": "alias",
        "target": "refs/heads/master",
        "expected_symlink": "refs/heads/master",
    },
    "refs/heads/alias-subdir": {
        "target_type": "alias",
        "target": "refs/heads/master",
        "expected_symlink": "master",
    },
    "refs/tags/alias-different-subdir": {
        "target_type": "alias",
        "target": "refs/heads/master",
        "expected_symlink": "../heads/master",
    },
}
# Origin
ORIGIN_URL = "https://github.com/rust-lang/rust"
ORIGIN_URL_ENCODED = "https%3A%2F%2Fgithub.com%2Frust-lang%2Frust"

# Special corner cases (not from Rust compiler)
REL_TARGET_CNT = "swh:1:rel:da5f9898d6248ab26277116f54aca855338401d2"
TARGET_CNT = "swh:1:cnt:be5effea679c057aec2bb020f0241b1d1d660840"
REL_TARGET_DIR = "swh:1:rel:3a7b2dfffed2945d2933ba4ebc063adba35ddb2e"
TARGET_DIR = "swh:1:dir:b24d39c928b9c3f440f8e2ec06c78f43d28d87d6"

ALL_ENTRIES = [
    REGULAR_FILE,
    ROOT_DIR,
    CNT_SYMLINK,
    DIR_SYMLINK,
    *REV_SYMLINKS,
    DIR_WITH_CNT_SYMLINK,
    DIR_WITH_DIR_SYMLINK,
    DIR_WITH_REV_SYMLINK,
    ROOT_REV,
    REV_SMALL_HISTORY,
    ROOT_REL,
    REL_TARGET_CNT,
    REL_TARGET_DIR,
    ROOT_SNP,
    ROOT_SNP_MASTER_BRANCH,
]



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from pathlib import Path
import re
from stat import S_IFDIR, S_IFLNK, S_IFREG
from typing import Any, AsyncIterator, Dict, Optional, Pattern, Sequence, Union

# Avoid cycling import
Fuse = "Fuse"


class EntryMode(IntEnum):
    """ Default entry mode and permissions for the FUSE.

    The FUSE mount is always read-only, even if permissions contradict this
    statement (in a context of a directory, entries are listed with permissions
    taken from the archive).
    """

    RDONLY_FILE = S_IFREG | 0o444
    RDONLY_DIR = S_IFDIR | 0o555
    SYMLINK = S_IFLNK | 0o444


@dataclass
class FuseEntry:
    """ Main wrapper class to manipulate virtual FUSE entries

    Attributes:
        name: entry filename
        mode: entry permission mode
        fuse: internal reference to the main FUSE class
        inode: unique integer identifying the entry
    """

    name: str
    mode: int
    depth: int
    fuse: Fuse
    inode: int = field(init=False)
    file_info_attrs: Dict[str, Any] = field(init=False, default_factory=dict)

    def __post_init__(self):
        self.inode = self.fuse._alloc_inode(self)
        # By default, let the kernel cache previously accessed data
        self.file_info_attrs["keep_cache"] = True

    async def size(self) -> int:
        """ Return the size (in bytes) of an entry """

        raise NotImplementedError

    def get_relative_root_path(self) -> str:
        return "../" * (self.depth - 1)

    def create_child(self, constructor: Any, **kwargs) -> FuseEntry:
        return constructor(depth=self.depth + 1, fuse=self.fuse, **kwargs)


@dataclass
class FuseFileEntry(FuseEntry):
    """ FUSE virtual file entry """

    async def get_content(self) -> bytes:
        """ Return the content of a file entry """

        raise NotImplementedError

    async def size(self) -> int:
        return len(await self.get_content())


@dataclass
class FuseDirEntry(FuseEntry):
    """ FUSE virtual directory entry """

    ENTRIES_REGEXP: Optional[Pattern] = field(init=False, default=None)

    async def size(self) -> int:
        return 0

    def validate_entry(self, name: str) -> bool:
        """ Return true if the name matches the directory entries regular
        expression, and false otherwise """

        if self.ENTRIES_REGEXP:
            return re.match(self.ENTRIES_REGEXP, name)
        else:
            return True

    async def compute_entries(self) -> Sequence[FuseEntry]:
        """ Return the child entries of a directory entry """

        raise NotImplementedError

    async def get_entries(self, offset: int = 0) -> AsyncIterator[FuseEntry]:
        """ Return the child entries of a directory entry using direntry cache """

        cache = self.fuse.cache.direntry.get(self)
        if cache:
            entries = cache
        else:
            entries = [x async for x in self.compute_entries()]
            self.fuse.cache.direntry.set(self, entries)

        # Avoid copy by manual iteration (instead of slicing) and use of a
        # generator (instead of returning the full list every time)
        for i in range(offset, len(entries)):
            yield entries[i]

    async def lookup(self, name: str) -> FuseEntry:
        """ Look up a FUSE entry by name """

        async for entry in self.get_entries():
            if entry.name == name:
                return entry
        return None


@dataclass
class FuseSymlinkEntry(FuseEntry):
    """ FUSE virtual symlink entry

    Attributes:
        target: path to symlink target
    """

    mode: int = field(init=False, default=int(EntryMode.SYMLINK))
    target: Union[str, bytes, Path]

    async def size(self) -> int:
        return len(str(self.target))

    def get_target(self) -> Union[str, bytes, Path]:
        """ Return the path target of a symlink entry """

        return self.target
